#include "MQPlugin.h"
#include "MQBasePlugin.h"

#include <unordered_map>

#ifndef _PluginDefine
#define _PluginDefine

class MorphCheckerWindow;

class CMQBaseObject : public cmq::CMQObject
{
public:
	CMQBaseObject(const cmq::CMQObject &obj);
	virtual ~CMQBaseObject() {}
	CMQBaseObject &operator=(const CMQBaseObject &other);

	const std::wstring &getObjName() const {return m_objname;}
	void setvisible(bool visible) {m_visible = visible;}
	bool visible() {return m_visible;}
	void hasMorph(bool morph) {m_hasmorph= morph;}
	bool hasMorph() {return m_hasmorph;}
private:
	std::wstring m_objname;
	bool m_hasmorph;
	bool m_visible;
};

class CMQViewObject : public CMQBaseObject
{
public:
	CMQViewObject(const cmq::CMQObject &obj) : CMQBaseObject(obj), m_morphed(false) {}
	virtual ~CMQViewObject() {}
	CMQViewObject &operator=(const CMQViewObject &other);

	void morphed() {m_morphed = true;}
	bool ismorphed() {return m_morphed;}
private:
	bool m_morphed;
};

class CMQMorphObject : public cmq::CMQObject
{
public:
	CMQMorphObject(const cmq::CMQObject &obj);
	virtual ~CMQMorphObject() {}
	CMQMorphObject &operator=(const CMQMorphObject &other);

	const std::wstring &getObjName() const {return m_objname;}
	const std::wstring &getMorphName() const {return m_morphname;}
	void setvalue(double value) {m_value = value; m_morphed = true;}
	double value() const {return m_value;}
	void setBase(UINT uid) {m_baseuid = uid;}
	UINT getBase() {return m_baseuid;}
//	void morphed() {m_morphed = true;}
	bool ismorphed() {return m_morphed;} // {return m_value != 0.0;}
private:
	std::wstring m_morphname;
	std::wstring m_objname;
	double m_value;
	UINT m_baseuid;
	bool m_morphed;
};

// CNX
class MorphChecker : public MQCommandPlugin
{
public:
	friend class MorphCheckerWindow;
	friend class MorphCheckerObjectListWindow;

	MorphChecker();
	virtual ~MorphChecker();

	virtual void GetPlugInID(DWORD *Product, DWORD *ID);
	virtual const char *GetPlugInName(void);
	virtual const char *EnumString(void);
	virtual BOOL Initialize();
	virtual void Exit();
	virtual BOOL Activate(MQDocument doc, BOOL flag);
	virtual BOOL IsActivated(MQDocument doc);
	void loadSetting();
	void saveSetting();

	virtual void OnNewDocument(MQDocument doc, const char *filename, NEW_DOCUMENT_PARAM& param);
	virtual void OnEndDocument(MQDocument doc);
	virtual void OnDraw(MQDocument doc, MQScene scene, int width, int height);
	virtual void OnUpdateObjectList(MQDocument doc);
	virtual BOOL OnUndo(MQDocument doc, int undo_state);
	virtual BOOL OnRedo(MQDocument doc, int undo_state);
	virtual BOOL OnMouseMove(MQDocument doc, MQScene scene, MOUSE_BUTTON_STATE& state);
	virtual BOOL OnLeftButtonDown(MQDocument doc, MQScene scene, MOUSE_BUTTON_STATE& state);
	virtual BOOL OnLeftButtonMove(MQDocument doc, MQScene scene, MOUSE_BUTTON_STATE& state);

	typedef bool (MorphChecker::*ExecuteCallbackProc)(MQDocument doc);

	void Execute(ExecuteCallbackProc proc);

	void set_morph(UINT uid, double value);
	double get_morph_value(UINT uid);
	void set_visible(UINT uid, bool value);
	bool get_visible(UINT uid);
	void set_editmode(bool mode) {m_editmode = mode;}
	void setSelectingMorph(UINT uid);

private:
	MorphCheckerWindow *m_Window;
	bool m_bActivate;
	MQDocument m_doc;
	bool m_objupdate;
	bool m_morphupdate;
	bool m_editmode;
	DWORD m_visibility;

	int posX;
	int posY;
	int height;
	int width;

	std::vector<CMQBaseObject> m_baseobjlist;
	std::vector<CMQMorphObject> m_morphobjlist;
	std::unordered_map<UINT, CMQViewObject> m_viewobjlist;

	// ԁAbNԕۑp
	std::unordered_map<cmq::oindex_t, bool> m_objvisible;
	std::unordered_map<cmq::oindex_t, bool> m_objlock;

	// _Ip
	std::vector<MQObject> m_select_viewobjlist;
	CMQBaseObject *m_select_base;
	CMQMorphObject *m_select_morph;
	cmq::vindex_t m_select_vindex;
	cmq::vindex_t m_select_vindex_symmetry; // I𒸓_̑Ώے_
	MQPoint m_select_mpoint0; // I𒸓_̃[tl0̏ꍇ̍W
	MQPoint m_select_mpoint100; // I𒸓_̃[tl100̏ꍇ̍W

	void make_objectlist(MQDocument doc_);
	void createDrawObject(MQDocument doc_, MQScene scene);
	void delete_viewobject(MQDocument doc);
	void morph(MQScene scene);
	MQPoint morph_one(MQPoint &vpoint0, MQPoint &bpoint, cmq::vindex_t vindex, std::vector<CMQMorphObject*> &morphobjsublist, CMQMorphObject *except = nullptr, bool full = false);

	UINT getBaseUIDfromMorphUID(MQDocument doc_, UINT morphuid);

};

#endif
